/***************************************************************************
                          kohonennet.cpp  -  description
                             -------------------
    begin                : Sun Jan 9 2000
    copyright            : (C) 2000 by Gavin James Wood
    email                : gav@indigoarchive.net
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#pragma implementation

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cmath>

using namespace std;

#include "kohonendata.h"
#include "kohonennet.h"

using namespace sucs;

unsigned Random(unsigned Low, unsigned High)
{
	return (rand() % (High - Low)) + Low;
}

double Random(double Low, double High)
{
	return (double(High - Low) * double(rand()) / double(RAND_MAX)) + double(Low);
}

void Shuffle(unsigned Ordinals[], unsigned Size)
{
	for(unsigned i = 0; i < Size * 2; i++)
	{	unsigned a = Random(0, Size), b = Random(0, Size);
		Ordinals[a] ^= Ordinals[b] ^= Ordinals[a] ^= Ordinals[b];
	}
}

void InitRandom()
{
	ifstream fin("/dev/random");
	if(fin)
	{
		unsigned a = fin.get(), b = fin.get(), c = fin.get();
		srand(a + b * 256 + c * 65536);
		fin.close();
	}
}

void KohonenNet::readCod(istream &fin)
{
	unsigned nWidth, nHeight, Dims;
	fin >> Dims >> nWidth >> nHeight;

	Deallocate();
	Allocate(nWidth, nHeight, Dims);

	char Buffer[1024];
	for(unsigned i = 0; !fin.eof() && i < (nWidth * nHeight); i++)
	{	for(unsigned d = 0; d < Dims; d++)
			fin >> theValues[i / nHeight][i % nHeight][d];
		while(fin.peek() == '\t' || fin.peek() == ' ') fin.get();
		fin.getline(Buffer, 1024);
		theLabels[i / nHeight][i % nHeight] = Buffer;
	}

}

void KohonenNet::writeCod(ostream &fout)
{
	fout << theDimensions << " " << theWidth << " " << theHeight << endl;
	for(unsigned x = 0; x < theWidth; x++)
		for(unsigned y = 0; y < theHeight; y++)
		{	for(unsigned d = 0; d < theDimensions; d++)
				fout << theValues[x][y][d] << " ";
			fout << theLabels[x][y] << endl;
		}
}

bool KohonenNet::DoTraining(unsigned RunLength, double Alpha, double Diameter, double Lambda, const KohonenData &Data)
{
	Train(RunLength, Alpha, Diameter, Lambda, Lambda, Data);
}

bool KohonenNet::Train(unsigned RunLength, double Alpha, double Diameter, double LambdaA, double LambdaD, const KohonenData &Data)
{
	if(theDimensions != Data.getDimensions()) return false;
	if(!RunLength) return false;
	unsigned Ordinals[Data.getCount()];
	for(unsigned i = 0; i < Data.getCount(); i++)
		Ordinals[i] = i;
	Shuffle(Ordinals, Data.getCount());
	for(unsigned i = 0; i < RunLength; i++)
	{	if(!(i % Data.getCount()) && i / Data.getCount() > 1)
		{	Diameter *= LambdaD;
			Alpha *= LambdaA;
			Shuffle(Ordinals, Data.getCount());
		}
		update(Data[Ordinals[i % Data.getCount()]], Alpha, Diameter);
	}
	return true;
}
	
bool KohonenNet::Train(unsigned RunLength, double Alpha, double Diameter, double K, const KohonenData &Data)
{
	if(theDimensions != Data.getDimensions()) return false;
	if(!RunLength) return false;

	double d = Diameter, a = Alpha, lambda = exp(-K);
	unsigned iterations = RunLength;
	
	unsigned curItem;
	for(curItem = 0; curItem < Data.getCount() && iterations >= 0; curItem++, iterations--)
		update(Data[curItem], a, d);

	unsigned index = 0;
	for(unsigned t = 0; t < iterations; t++, curItem++)
	{
		index = (index + 1) % Data.getCount();
		if(curItem >= Data.getCount())
		{	curItem = 0;
			d *= lambda;
			a *= lambda;
		}
		update(Data[index], a, d);
	}
	return true;
}

string KohonenNet::Classify(const double *Vector) const
{
	unsigned x, y;
	closest(Vector, x, y);
	return theLabels[x][y];
}

void KohonenNet::Label(double High, double Low, const KohonenData &Data)
{
	// Generate a list of the labels (not can be more than data entry per label)
	vector<string> theUniqueLabels;
	for(unsigned i = 0; i < Data.getCount(); i++)
	{	vector<string>::iterator j;
		for(j = theUniqueLabels.begin(); j != theUniqueLabels.end(); j++)
			if(*j == Data.getLabel(i)) break;
		if(j == theUniqueLabels.end())
			theUniqueLabels.push_back(Data.getLabel(i));
	}

	// Generate a stats matrix
	unsigned Stats[theWidth][theHeight][theUniqueLabels.size()];
	for(unsigned x = 0; x < theWidth; x++)
		for(unsigned y = 0; y < theHeight; y++)
			for(unsigned l = 0; l < theUniqueLabels.size(); l++)
				Stats[x][y][l] = 0;
	for(unsigned i = 0; i < Data.getCount(); i++)
	{	unsigned l = 0;
		for(vector<string>::iterator j = theUniqueLabels.begin(); j != theUniqueLabels.end(); j++, l++)
			if(*j == Data.getLabel(i))
			{	unsigned x, y;
				closest(Data[i], x, y);
				Stats[x][y][l]++;
				break;
			}
	}

	// Use stats to label nodes
	for(unsigned x = 0; x < theWidth; x++)
		for(unsigned y = 0; y < theHeight; y++)
		{	unsigned *ratings = Stats[x][y];
			double First = 0, Second = 0, Sum = 0;
			string Winner;
			unsigned l = 0;
			for(vector<string>::iterator j = theUniqueLabels.begin(); j != theUniqueLabels.end(); j++, l++)
			{	Sum += ratings[l];
				if(ratings[l] >= First)
				{	Second = First;
					First = ratings[l];
					Winner = *j;
				}
				else if(ratings[l] >= Second)
					Second = ratings[l];
			}	
			if(First >= High * Sum && Second < Low * Sum)
				theLabels[x][y] = Winner;
			else
				theLabels[x][y] = "";
		}
}

void KohonenNet::update(const double *Vector, double Alpha, double Diameter)
{
	unsigned closestX, closestY;
    closest(Vector, closestX, closestY);
	for(unsigned x = 0; x < theWidth; x++)
		for(unsigned y = 0; y < theHeight; y++)
		{	double p = proximity(closestX, closestY, x, y, Diameter);
			if(p < 0.0001) continue;
			for(unsigned i = 0; i < theDimensions; i++)
				theValues[x][y][i] += Alpha * p * (Vector[i] - theValues[x][y][i]);
		}
}

double KohonenNet::closest(const double *Vector, unsigned &returnX, unsigned &returnY) const
{
    double minDist = 1.0e90;
	for(unsigned x = 0; x < theWidth; x++)
		for(unsigned y = 0; y < theHeight; y++)
		{	double Distance = 0;
			for(unsigned d = 0; d < theDimensions; d++)
			{	double thisDist = Vector[d] - theValues[x][y][d];
				Distance += thisDist * thisDist;
			}
			if(Distance < minDist)
			{	minDist = Distance;
				returnX = x;
				returnY = y;
			}
		}
	return minDist;
}

double KohonenNet::proximity(unsigned x1, unsigned y1, unsigned x2, unsigned y2, double Diameter) const
{
    return ((abs((signed)(x1 - x2)) <= Diameter) && (abs((signed)(y1 - y2)) <= Diameter)) ? 1 : 0;
}

KohonenNet::KohonenNet()
{
	theValues = 0;
	theLabels = 0;
	theWidth = 0;
	theHeight = 0;
	theDimensions = 0;
}

KohonenNet::~KohonenNet()
{
	Deallocate();
}

void KohonenNet::Init(unsigned Width, unsigned Height, unsigned Dimensions, double Low, double High)
{
	Deallocate();
	Allocate(Width, Height, Dimensions);
	InitRandom();

	for(unsigned x = 0; x < theWidth; x++)
		for(unsigned y = 0; y < theHeight; y++)
		{	for(unsigned d = 0; d < theDimensions; d++)
				theValues[x][y][d] = Random(Low, High);
			theLabels[x][y] = "";
		}
}

void KohonenNet::Allocate(unsigned Width, unsigned Height, unsigned Dimensions)
{
	theValues = new double **[Width];
	theLabels = new string *[Width];
	for(unsigned x = 0; x < Width; x++)
	{	theValues[x] = new double *[Height];
		theLabels[x] = new string[Height];
		for(unsigned y = 0; y < Height; y++)
			theValues[x][y] = new double[Dimensions];
	}
	theWidth = Width;
	theHeight = Height;
	theDimensions = Dimensions;
}

void KohonenNet::Deallocate()
{
	if(theValues || theLabels)
	{	if(!(theValues && theLabels)) exit(1);

		for(unsigned x = 0; x < theWidth; x++)
		{	for(unsigned y = 0; y < theHeight; y++)
				delete [] theValues[x][y];
			delete [] theLabels[x];
			delete [] theValues[x];
		}
		delete [] theValues;
		delete [] theLabels;
	}
	theValues = 0;
	theLabels = 0;
	theWidth = 0;
	theHeight = 0;
	theDimensions = 0;
}

